import json
from time import monotonic, sleep
from adafruit_clue import clue
from adafruit_ble import BLERadio
from adafruit_ble.advertising.standard import ProvideServicesAdvertisement
from adafruit_ble.services.nordic import UARTService
from gui import Gui, SAND_COLOR_OUT_OF_TIME
from splash import Splash
from math import floor
from debug import debug

# Things you may want to customize
DEVICE_NAME = "Clue"                # Name of device when listing Bluetooth devices
GESTURE_TYPE = "None"              # 'Cover' or 'Direction'. Anything else turns both off
COVER_GESTURE_THRESHOLD_ON = 180    # How close do you have to be to trigger cover gesture?
COVER_GESTURE_THRESHOLD_OFF = 20    # How far   do you have to be to trigger uncover gesture?
COVER_GESTURE_DELAY = 3             # How long you have to cover the sensor to trigger Covered
ENABLE_SWIPE_DETECTION = False      # The touchpads are easy to hit. Turn them off if not using it.
ENABLE_SHAKE_DETECTION = False      # Might slow down things a tiny bit. Enable if you like.
SHAKE_THRESHOLD = 20                # How sensitive should it be for detecting shakes?
DOUBLE_SHAKE_PREVENTION_TIMEOUT = 3 # Seconds after a shake when no second shake can be detected

ORIENTATION_CHANGE_DELAY = 1        # How long to debounce changes in orientation
ORIENTATION_ON_THRESHOLD = 9.5      # 0-10. Higher number, less sensitive.
ORIENTATION_OFF_THRESHOLD = 4.0     # 0-10. Higher number, more sensitive.
ORIENTATION_RIGHT = 'Right'         # The names of the various directions.
ORIENTATION_LEFT = 'Left'           # These are sent as 'To/From <name>'
ORIENTATION_STANDING = 'Standing'
ORIENTATION_UPSIDE_DOWN = 'Upside Down'
ORIENTATION_FACE_DOWN = 'Face Down'
ORIENTATION_FACE_UP = 'Face Up'
ORIENTATION_ACTIVATIONS = {         # Which orientations should you send to the SGT?
    ORIENTATION_RIGHT: False,       # Enabling all should work, but will send a lot of
    ORIENTATION_LEFT: False,        # notifications to the timer.
    ORIENTATION_STANDING: False,
    ORIENTATION_UPSIDE_DOWN: True,
    ORIENTATION_FACE_DOWN: True,
    ORIENTATION_FACE_UP: False,
}
BRIGHTENESS_OPTIONS = [0.06, 0.2, 0.4, 0.7, 1.0]    # The different brightnesses to cycle through
BRIGHTENESS_INITIAL_INDEX = 3       # 0-indexed. Which brightness to start with.

# Suggested Script and Action Mapping
# These are sent on connection to the SGT to pre-populate the Action/Write scripts for quick save.
suggestions = {
    "script": [
        '0 sgtTimerMode;sgtState;sgtColor;sgtTurnTime;sgtName;sgtPlayerTime;sgtTotalPlayTime%0A'
    ],
    "scriptName": DEVICE_NAME + " Write",
    "actionMap": [
        ('Button AB', 'remoteActionUndo'),
        ('Button A', 'remoteActionPrimary'),
        ('Button B', 'remoteActionSecondary'),
        ('Connected', 'remoteActionPoll'),
        ('To Upside Down', 'remoteActionTurnAdminOn'),
        ('From Upside Down', 'remoteActionTurnAdminOff'),
        ('To Face Down', 'remoteActionTurnPauseOn'),
        ('From Face Down', 'remoteActionTurnPauseOff'),
    ],
    "actionMapName": DEVICE_NAME + " Actions",
}

################## END OF THINGS YOU SHOULD MESS WITH ################

# Constants
STATE_PLAYING = 'pl'
STATE_ADMIN = 'ad'
STATE_PAUSE = 'pa'
STATE_START = 'st'
STATE_FINISHED = 'en'
STATE_NOT_CONNECTED = 'nc'
STATE_RUNNING = 'ru'
STATE_NOT_RUNNING = 'nr'

TIMER_MODE_COUNT_UP = 'cu'
TIMER_MODE_COUNT_DOWN = 'cd'
TIMER_MODE_SAND_TIMER = 'st'
TIMER_MODE_NO_TIMER = 'nt'


# State Setup
sgt_timer_mode = TIMER_MODE_COUNT_UP # Current timer-mode (cd/cu/st/nt for Count-Down/Up, SandTimer, No Timer)
sgt_state=STATE_NOT_CONNECTED        # The current state. SandTimer,
                                     # Sand, ru/nr/pa/en for running, not running, paused or end
                                     # Not Sand, st/en/pa/ad/pl for start, end, pause, admin or playing
sgt_color = None                # (not sand) The current or next-up player color
sgt_name=""                     # (not sand) The current or next-up player name

sgt_turn_time_sec=0             # Count-Up, time taken this turn or pause time or admin time
                                # Count-Down, same as above, but negative values during Delay Time
                                # Sand, time taken out of the sand timer

sgt_player_time_sec=0           # Count-Up, total time taken, or blank for admin/pause time
                                # Count-Down, remaining time bank, or blank for admin/pause time
                                # Sand, sand-timer reset size

sgt_total_play_time_sec=0       # Count-Up/Down, total play time, not counting this turn and not admin/pause time

sgt_update_ts=monotonic()       # When did we last get an update from SGT?
sgt_last_shake_ts=0             # When did we last shake the playground?
sgt_incomplete_line_read=''     # If we've read data not ending in new line, store it here for now
sgt_last_read_line=''           # Last completed line read. Used to avoid duplicate lines.
sgt_last_screen_update_ts = 0   # Used to slow down screen updates to once a second
sgt_is_covered = False          # Keep track of if we are currently covering the proximity sensor
sgt_covered_ts = None           # When was the sensor covered?
sgt_orientation = None          # Up, Down, Front, Back, Left or Right (string)
sgt_orientation_tmp = None      # What is the latest detected orientation
sgt_orientation_tmp_ts = 0      # When did we set the sgt_orientation_tmp value.


# Display Setup
clue.display.auto_brightness = False
clue.display.brightness = 1
clue.display.rotation = 0
current_brightness_index = BRIGHTENESS_INITIAL_INDEX
clue.display.brightness = BRIGHTENESS_OPTIONS[current_brightness_index]
Splash.show()
Gui.load_fonts()

# Bluetooth Setup
ble = BLERadio()
uart = UARTService()
advertisement = ProvideServicesAdvertisement(uart)

""" Read from the UART TX channel and update the global state if new info came in. """
def read_state():
    global sgt_incomplete_line_read
    while uart.in_waiting > 0:
        text_to_process = sgt_incomplete_line_read + str(uart.read(uart.in_waiting), 'utf-8')
        lines = text_to_process.split("\n")
        if len(lines) == 0:
            return
        last_item = lines.pop()
        if last_item == '':
            do_this_line = lines.pop()
            if len(lines) > 0:
                debug(f"SKIP: {lines}")
            sgt_incomplete_line_read = ''
            sleep(0.1)
            if uart.in_waiting:
                debug(f"SKIP AFTER ALL: {do_this_line}")
            else:
                handle_state_update(do_this_line)
        else:
            if len(lines) > 0:
                print(f"SKIP: {lines}")
            debug(f"Incomplete Line: {last_item}")
            sgt_incomplete_line_read = last_item

def handle_state_update(state_line):
    if state_line == "GET SETUP SUGGESTIONS":
        send(json.dumps(suggestions))
        send("Connected")
        return
    if (sgt_last_read_line == state_line):
        return
    debug(f"Handle State Update: {state_line}")

    global sgt_timer_mode, sgt_state, sgt_color, sgt_turn_time_sec, sgt_update_ts, sgt_name, sgt_player_time_sec, sgt_total_play_time_sec, sgt_last_screen_update_ts
    sgt_last_screen_update_ts = 0
    try:
        tmp_timer_mode, tmp_state, color_hex, tmp_turn_time, sgt_name, tmp_player_time_sec, tmp_total_play_time_sec = [item.strip() for item in state_line.split(";")]
        sgt_turn_time_sec = int(tmp_turn_time) if len(tmp_turn_time) > 0 else 0
        sgt_player_time_sec = int(tmp_player_time_sec) if len(tmp_player_time_sec) > 0 else 0
        sgt_total_play_time_sec = int(tmp_total_play_time_sec) if len(tmp_total_play_time_sec) > 0 else 0
        sgt_color = simplify_color((int(color_hex[0:2],16),int(color_hex[2:4],16),int(color_hex[4:6],16))) if len(color_hex) == 6 else (255,255,255)
        sgt_update_ts=monotonic()

        if (tmp_timer_mode != sgt_timer_mode or sgt_state != tmp_state):
            sgt_timer_mode = tmp_timer_mode
            sgt_state = tmp_state

            if (sgt_state == STATE_FINISHED or sgt_timer_mode == TIMER_MODE_NO_TIMER):
                switch_to_titles_only_gui()
            else:
                switch_to_titles_and_times_gui()

        if sgt_state == STATE_PLAYING:
            switch_to_playing()
        elif sgt_state == STATE_ADMIN:
            switch_to_admin_time()
        elif sgt_state == STATE_PAUSE:
            switch_to_paused()
        elif sgt_state == STATE_FINISHED:
            switch_to_end()
        elif sgt_state == STATE_START:
            switch_to_start()
        elif sgt_state == STATE_RUNNING:
            switch_to_sandtimer_running()
        elif sgt_state == STATE_NOT_RUNNING:
            switch_to_sandtimer_not_running()
        else:
            switch_to_error()
    except Exception as e:
        result = repr(e)
        print(result)

def simplify_color(color):
    return (simplify_color_part(color[0]), simplify_color_part(color[1]), simplify_color_part(color[2]))

def simplify_color_part(color_part):
    return (color_part//86)*127

def switch_to_titles_only_gui():
    debug("Switch to TITLES ONLY")
    Splash.hide()
    Gui.show_titles_only()

def switch_to_titles_and_times_gui():
    debug("Switch to TITLES AND TIMES")
    Splash.hide()
    Gui.show_titles_and_times()

def switch_to_playing():
    debug("Switch to PLAYING")
    Gui.set_title(sgt_name)
    Gui.set_background(sgt_color)

def switch_to_admin_time():
    debug("Switch to Admin Time")
    Gui.set_title("Admin Time", sgt_name)
    Gui.set_total_time("")
    Gui.set_background(sgt_color)

def switch_to_paused():
    debug("Switch to Paused")
    Gui.set_title("Paused", sgt_name)
    Gui.set_total_time("")
    Gui.set_background(sgt_color)

def switch_to_start():
    debug("Switch to Start")
    Gui.set_turn_time("")
    Gui.set_total_time("")
    Gui.set_title('Starting', sgt_name)
    Gui.set_background(sgt_color)

def switch_to_end():
    debug("Switch to End")
    global sgt_color
    sgt_color = (0,0,0)
    Gui.set_title("Game Over")
    Gui.set_background(sgt_color)

def switch_to_sandtimer_running():
    debug("Switch to Sand Timer Running")
    Gui.set_title('', sec_to_text(sgt_player_time_sec))

def switch_to_sandtimer_not_running():
    debug("Switch to Sand Timer Not Running")
    Gui.set_title('', sec_to_text(sgt_player_time_sec))

def switch_to_trying_to_connect():
    global sgt_state
    sgt_state = STATE_NOT_CONNECTED
    debug("Switch to Trying to Connect")
    Splash.show()

def switch_to_connecting():
    global sgt_state
    sgt_state = STATE_NOT_CONNECTED
    debug("Switch to Connecting")
    Splash.show()

def switch_to_error():
    debug("Switch to Error")
    Gui.set_turn_time("")
    Gui.set_total_time("")
    Gui.set_title("Error")
    Gui.set_background(sgt_color)

def change_brightness():
    global current_brightness_index
    current_brightness_index = (current_brightness_index + 1) % len(BRIGHTENESS_OPTIONS)
    clue.display.brightness = BRIGHTENESS_OPTIONS[current_brightness_index]

def check_buttons():
    if sgt_orientation == ORIENTATION_FACE_DOWN or sgt_orientation_tmp == ORIENTATION_FACE_DOWN:
        return
    if clue.button_a:
        clue.red_led = True
        double_press = False
        while clue.button_a:
            if clue.button_b:
                double_press = True
                while clue.button_b:
                    update_screen()
            update_screen()
        if double_press:
            send("Button AB")
        else:
            send("Button A")
        clue.red_led = False
    elif clue.button_b:
        clue.red_led = True
        double_press = False
        while clue.button_b:
            if clue.button_a:
                double_press = True
                while clue.button_a:
                    update_screen()
            update_screen()
        if double_press:
            send("Button AB")
        else:
            send("Button B")
        clue.red_led = False
    elif ENABLE_SWIPE_DETECTION and clue.touch_0:
        now = monotonic()
        while (monotonic()-now < 2 and not clue.touch_1):
            pass
        while (monotonic()-now < 2 and not clue.touch_2):
            pass
        if (monotonic()-now < 2):
            send("Swipe LTR")
    elif ENABLE_SWIPE_DETECTION and clue.touch_2:
        now = monotonic()
        while (monotonic()-now < 2 and not clue.touch_1):
            pass
        while (monotonic()-now < 2 and not clue.touch_0):
            pass
        if (monotonic()-now < 2):
            send("Swipe RTL")
    elif clue.touch_1:
        change_brightness()
        while (clue.touch_1):
            pass

def check_gesture_cover():
    global sgt_is_covered, sgt_covered_ts
    if sgt_orientation == ORIENTATION_FACE_DOWN or sgt_orientation_tmp == ORIENTATION_FACE_DOWN:
        return
    if (sgt_is_covered and clue.proximity < COVER_GESTURE_THRESHOLD_OFF):
        sgt_is_covered = False
        sgt_covered_ts = None
        send("Uncovered")
    elif (not sgt_is_covered and clue.proximity > COVER_GESTURE_THRESHOLD_ON and sgt_covered_ts == None):
        sgt_covered_ts = monotonic()
    elif (not sgt_is_covered and clue.proximity > COVER_GESTURE_THRESHOLD_ON and monotonic() - sgt_covered_ts > COVER_GESTURE_DELAY):
        send("Covered")
        sgt_is_covered = True


def check_gesture_direction():
    if sgt_orientation == ORIENTATION_FACE_DOWN or sgt_orientation_tmp == ORIENTATION_FACE_DOWN:
        return
    value = clue.gesture
    if value == 1:
        send("Gesture Up")
    elif value == 2:
        send("Gesture Down")
    elif value == 3:
        send("Gesture Left")
    elif value == 4:
        send("Gesture Right")

def orient_screen():
    check_rotation_screen_axis(clue.acceleration[0], 270, 90, clue.acceleration[1], clue.acceleration[2])
    check_rotation_screen_axis(clue.acceleration[1], 0, 180, clue.acceleration[0], clue.acceleration[2])
    check_rotation_screen_axis(clue.acceleration[2], 0, 0, clue.acceleration[0], clue.acceleration[1])

def check_rotation_screen_axis(value, pos_rotation, neg_rotation, other_axis_one, other_axis_two):
    if (abs(other_axis_one) > ORIENTATION_OFF_THRESHOLD):
         return
    if (abs(other_axis_two) > ORIENTATION_OFF_THRESHOLD):
        return
    if (clue.display.rotation != pos_rotation):
        if value > ORIENTATION_ON_THRESHOLD:
            clue.display.rotation = pos_rotation
    if (clue.display.rotation != neg_rotation):
        if -value > ORIENTATION_ON_THRESHOLD:
            clue.display.rotation = neg_rotation

def check_orientation():
    check_orientation_axis(clue.acceleration[0], ORIENTATION_RIGHT, ORIENTATION_LEFT)
    check_orientation_axis(clue.acceleration[1], ORIENTATION_STANDING, ORIENTATION_UPSIDE_DOWN)
    check_orientation_axis(clue.acceleration[2], ORIENTATION_FACE_DOWN, ORIENTATION_FACE_UP)
    if sgt_orientation == ORIENTATION_FACE_DOWN:
        clue.display.brightness = 0
    else:
        clue.display.brightness = BRIGHTENESS_OPTIONS[current_brightness_index]

def check_orientation_axis(value, pos_orientation, neg_orientation):
    global sgt_orientation, sgt_orientation_tmp, sgt_orientation_tmp_ts
    if sgt_orientation_tmp == None:
        if ORIENTATION_ACTIVATIONS[pos_orientation] and value > ORIENTATION_ON_THRESHOLD:
            sgt_orientation_tmp = pos_orientation
            sgt_orientation_tmp_ts = monotonic()
        elif ORIENTATION_ACTIVATIONS[neg_orientation] and -value > ORIENTATION_ON_THRESHOLD:
            sgt_orientation_tmp = neg_orientation
            sgt_orientation_tmp_ts = monotonic()
    elif ORIENTATION_ACTIVATIONS[pos_orientation] and sgt_orientation_tmp == pos_orientation and value < ORIENTATION_OFF_THRESHOLD:
        sgt_orientation_tmp = None
        sgt_orientation_tmp_ts = monotonic()
    elif ORIENTATION_ACTIVATIONS[neg_orientation] and sgt_orientation_tmp == neg_orientation and -value < ORIENTATION_OFF_THRESHOLD:
        sgt_orientation_tmp = None
        sgt_orientation_tmp_ts = monotonic()

    if (sgt_orientation != sgt_orientation_tmp and monotonic() - sgt_orientation_tmp_ts > ORIENTATION_CHANGE_DELAY):
        if (sgt_orientation != None):
            send(f"From {sgt_orientation}")
        if (sgt_orientation_tmp != None):
            send(f"To {sgt_orientation_tmp}")
        sgt_orientation = sgt_orientation_tmp

def check_shake():
    global sgt_last_shake_ts
    if monotonic() - sgt_last_shake_ts > DOUBLE_SHAKE_PREVENTION_TIMEOUT:
        if clue.shake(shake_threshold=SHAKE_THRESHOLD):
            send("Shake")
            sgt_last_shake_ts = monotonic()

def sec_to_text(s):
    sec_abs = abs(s)
    t = 0
    if (sec_abs > 5460):
        t = '{} h, {} min'.format(floor(sec_abs / 3600), round((sec_abs % 3600) / 60))
    elif (sec_abs > 99):
        t = '{} min'.format(round(sec_abs / 60))
    else:
        t = '{} sec'.format(round(sec_abs))

    if (s < 0):
        return '({})'.format(t)
    else:
        return t

def update_screen():
    global sgt_color, sgt_last_screen_update_ts
    if ((monotonic() - sgt_last_screen_update_ts) < 1) :
            return

    sgt_last_screen_update_ts = monotonic()

    if sgt_timer_mode == TIMER_MODE_COUNT_UP:
        if sgt_state in [STATE_PLAYING, STATE_ADMIN, STATE_PAUSE]:
            time_added_by_monotonic = monotonic() - sgt_update_ts
            turn_time = sgt_turn_time_sec + time_added_by_monotonic
            Gui.set_turn_time(sec_to_text(turn_time))
            if sgt_state == STATE_PLAYING:
                percent = round(100 * (time_added_by_monotonic + sgt_player_time_sec) / (time_added_by_monotonic + sgt_total_play_time_sec))
                Gui.set_total_time('{}%'.format(percent))
        elif sgt_state in [STATE_START, STATE_FINISHED, STATE_NOT_CONNECTED]:
            pass
        else:
            switch_to_error()

    elif sgt_timer_mode == TIMER_MODE_COUNT_DOWN:
        if sgt_state in [STATE_PLAYING, STATE_ADMIN, STATE_PAUSE]:
            time_added_by_monotonic = monotonic() - sgt_update_ts
            turn_time = sgt_turn_time_sec + time_added_by_monotonic
            Gui.set_turn_time(sec_to_text(turn_time))
            if sgt_state == STATE_PLAYING:
                remaining_time_bank = sgt_player_time_sec
                if (sgt_turn_time_sec < 0):
                    remaining_time_bank -= max(0, time_added_by_monotonic + sgt_turn_time_sec)
                else:
                    remaining_time_bank -= time_added_by_monotonic
                Gui.set_total_time(sec_to_text(remaining_time_bank))
        elif sgt_state in [STATE_START, STATE_FINISHED, STATE_NOT_CONNECTED]:
            pass
        else:
            switch_to_error()

    elif sgt_timer_mode == TIMER_MODE_NO_TIMER:
        if sgt_state in [STATE_PLAYING, STATE_ADMIN, STATE_PAUSE, STATE_START, STATE_FINISHED, STATE_NOT_CONNECTED]:
            pass
        else:
            switch_to_error()

    elif sgt_timer_mode == TIMER_MODE_SAND_TIMER:
        if sgt_state == STATE_RUNNING:
            time_added_by_monotonic = monotonic() - sgt_update_ts
            turn_time = sgt_turn_time_sec + time_added_by_monotonic
            remaining_time = max(sgt_player_time_sec - turn_time, 0)
            if (remaining_time == 0):
                Gui.set_background(SAND_COLOR_OUT_OF_TIME)
                Gui.set_title('(still running)', sec_to_text(sgt_player_time_sec))
                Gui.set_turn_time("Out of Time!")
            else:
                Gui.update_sandtimer_position(remaining_time/ sgt_player_time_sec)
                Gui.set_turn_time(sec_to_text(remaining_time))
        elif sgt_state == STATE_NOT_RUNNING:
            remaining_time = max(sgt_player_time_sec - sgt_turn_time_sec, 0)
            if (remaining_time == 0):
                Gui.set_background(SAND_COLOR_OUT_OF_TIME)
                Gui.set_title('(stopped)', sec_to_text(sgt_player_time_sec))
                Gui.set_turn_time("Out of Time!")
            else:
                Gui.update_sandtimer_position(remaining_time/ sgt_player_time_sec)
                Gui.set_turn_time(sec_to_text(remaining_time))
        elif sgt_state == STATE_PAUSE:
            time_added_by_monotonic = monotonic() - sgt_update_ts
            turn_time = sgt_turn_time_sec + time_added_by_monotonic
            Gui.set_background((0, 0, 0))
            Gui.set_turn_time(sec_to_text(turn_time))
        elif sgt_state in [STATE_FINISHED, STATE_NOT_CONNECTED]:
            pass
        else:
            switch_to_error()
    else:
        switch_to_error()

def send(string):
    debug(string)
    uart.write((string+"\n").encode("utf-8"))

ble.name = DEVICE_NAME
while True:
    ble.start_advertising(advertisement)
    clue.display.rotation = 0
    debug("Waiting to connect")
    switch_to_trying_to_connect()
    update_screen()
    while not ble.connected:
        pass
    ble.stop_advertising()
    switch_to_connecting()
    ts = monotonic()
    while monotonic() - ts < 2:
        update_screen()
    send("Connected")
    while ble.connected:
        read_state()
        orient_screen()
        update_screen()
        check_buttons()
        if (GESTURE_TYPE == 'Cover'):
            check_gesture_cover()
        elif (GESTURE_TYPE == 'Direction'):
            check_gesture_direction()
        if (ENABLE_SHAKE_DETECTION):
            check_shake()
        check_orientation()
    debug('Disconnected!')
